' sqlite3 means verison 3 of the sqlite dll must be used
' it can be downloaded from: http://www.sqlite.org/download.html
' you want the: 
'    Precompiled Binaries for Windows
'    or
'    Precompiled Binaries for Linux
' The file taht you are looking for is sqlite3.dll
' It must be located somewhere you application can
' find it (system32 or your app directory)
'-------------------------------------------------------
#include once "C:\bin\FreeBASIC\inc\sqlite3.bi"


' -----SQLITE3 Object, filehandles, etc-------------------------------------
' Each open SQLite database is represented by a pointer to an instance of 
' the opaque structure named "sqlite3". It is useful to think of an sqlite3 
' pointer as an object.
'
' That is all the docs say but we need some of these. 
' We are going to allocate to a few of them along with
' a bit of tracking data
Type SQL3FileType
	DbName as String		' filename for DB
	Handle as sqlite3 ptr	' SQLite file handdle
	Status  as Byte			' 0 if unused, 1 open
End Type

Dim Shared SQL3Initialized as Byte = 0
Dim Shared SQL3FileLimit as Byte = 5
Dim Shared SQL3File () as SQL3FileType
Dim Shared SQL3TableCount as Integer
Dim Shared SQL3TableNames() as String
Dim Shared SQL3ColumnCount as Integer
Dim Shared SQL3ColumnNames() as String
Dim Shared SQL3ColumnTypes() as String
Dim Shared SQL3ColumnNull() as String
Dim Shared SQL3ColumnDefault() as String
Dim Shared SQL3ColumnPKey() as String
Dim Shared SQL3errMsg as zstring ptr    ' place for SQLite error message
Dim Shared SQL3errCode as Integer    ' place for SQLite error message
Dim Shared SQL3DataFields() as String

Dim Shared SQL3EOL as String            ' End of Line/Row
Dim Shared SQL3DLM as String            ' Field delimiter

'------------------------------------------------------
Private Sub SQLiteInitialize ()
	' sets values in structure to known values
    ' print "Init SQL3 structures"          ' debugging
    SQL3EOL=chr(13) & Chr(10)
    SQL3DLM=chr(9)
	Redim SQL3File (0 to SQL3FileLimit)
	Dim I as Integer
	for I=0 to SQL3FileLimit
		SQL3File(I).DbName=""
		SQL3File(I).Status=0
	next i
	SQL3Initialized=SQL3FileLimit
end sub

'------------------------------------------------------
Public Function SQL3_DBopen (DBName as String) as Integer
	' Basic open function
	' Returns negative number on failure
	' returns SQL3File file number (1 to 10) on sucess
	'-----------------------------------------
    ' Print "Open DB"           ' debugging
	if SQL3Initialized=0 then SQLiteInitialize
	' First see if we have a valid filename
	if dir(DBName)="" then return -1
	' now find an open SQLite structure
	Dim N as Integer
	N=1
	while (N<(SQL3FileLimit+1)) and (SQL3File(N).Status<>0)
		N=N+1
	wend
	if (N>SQL3FileLimit) then return -2
	SQL3errCode = sqlite3_open( DBName, @SQL3File(N).Handle ) 
    If SQL3errCode = SQLITE_OK then
		SQL3File(N).DbName=DBName
		SQL3File(N).Status=1
		Return N
	Else
        ' basically all this does is deallocate memory
		sqlite3_close(SQL3File(N).Handle) 
		Return -3
	end if 
End Function

'------------------------------------------------------
Public Function SQL3_DBclose (FileNumber as Integer) as Integer
	' close the SQLite3 DB according to the SQLite file numnber
	' return zero on success
	if SQL3File(FileNumber).DbName="" then return -1
	if dir(SQL3File(FileNumber).DbName)="" then return -2
	' if it is closed return a zero
	if SQL3File(FileNumber).Status=0 then return 0
	Dim X as Integer
	Dim N as Integer
	N=0
	X=SQLITE_BUSY				' SQLITE_BUSY is defined to be 5
	While X=SQLITE_BUSY
		SQL3errCode=sqlite3_close(SQL3File(FileNumber).Handle)
		N=N+1
		If SQL3errCode=SQLITE_BUSY then Sleep (25)
        X=SQL3errCode
		If N>100 then 
			' used a defered closure, this can lead to a memory leak
			SQL3errCode=sqlite3_close_v2(SQL3File(FileNumber).Handle)
			X=0
		end if
	Wend
	SQL3File(FileNumber).Status=0
	Return 0
End Function

'------------------------------------------------------
Sub SQL3_DBcloseAll () 
	' close all the SQLite3 Databases
	Dim I as Integer
	Dim X as Integer
	For I=1 to SQL3FileLimit
		' Note that we are ignoring the return code
		' This is intended to be used before a program exit
		' and we figure the user is going to eixt
		' regardless of the status
		if (SQL3File(I).Status<>0) then X=SQL3_DBclose(I)
	Next I
End SUb

'------------------------------------------------------
' open and close were the easy parts --- 
' now we need to be able to get basic information about the database
'------------------------------------------------------
' first some help with checking for errors 
Public Function SQL3_ErrMsg (FileNumber as Integer) as String
    ' get the most recent error message
    ' const char *sqlite3_errmsg(sqlite3*) 
    Dim S as String
    S = *sqlite3_errmsg(SQL3File(FileNumber).Handle)
    If S="not an error" then
        return ""
    else
        return S
    end if
End Function

Public Function SQL3_ErrCode (FileNumber as Integer) as Integer
    ' get the most recent extended error code
    ' int sqlite3_extended_errcode(sqlite3 *db);
    SQL3_ErrCode = sqlite3_extended_errcode(SQL3File(FileNumber).Handle)
End Function

'------------------------------------------------------
'  Return the number of columns for a given SQL Query
Private Function SQL3QueryColCountCallback cdecl ( _
	byval Cols as any ptr, _              ' 
	byval ColCount as integer, _            ' number of columns
	byval RowData as zstring ptr ptr, _     ' string array of data for one row
	byval ColName as zstring ptr ptr _      ' string array of column names
    ) as integer
    Poke Long, Cols,ColCount
    return -1                               ' abort query, we have what we want
end function

Public Function SQL3QueryColCount (FileNumber as Integer, Sql as String) as Long
    ' returns the number of row for a SQL Query
    ' negative number indicates an error
    Dim Cols As Long
    SQL3errCode= sqlite3_exec( SQL3File(FileNumber).Handle, SQL, _
                     @SQL3QueryColCountCallback, @Cols, @SQL3errMsg)
    If SQL3errCode=4 then SQL3errCode=0
    Return Cols
End Function
'------------------------------------------------------
' Plan "B", Cycle quickly through the entire thing
'------------------------------------------------------
'  Return the number of Rows for a given SQL Query
Private Function SQL3QueryRowCountCallback cdecl ( _
	byval Rows as any ptr, _              ' 
	byval ColCount as integer, _            ' number of columns
	byval RowData as zstring ptr ptr, _     ' string array of data for one row
	byval ColName as zstring ptr ptr _      ' string array of column names
    ) as integer
    ' This should only return one row with on column that contains the rowcount
    Poke Long, Rows, Peek(Long,Rows)+1
    return 0
end function

Public Function SQL3QueryRowCount (FileNumber as Integer, Sql as String) as Long
    ' returns the number of row for a SQL Query
    ' negative number indicates an error
    Dim Rows As Long
    SQL3errCode= sqlite3_exec( SQL3File(FileNumber).Handle, SQL, _ 
                     @SQL3QueryRowCountCallback, @Rows, @SQL3errMsg)
    Return Rows
End Function
'------------------------------------------------------
' Now for table names --- 
'
' This function fills the array
' "SQL3TableNames" with the names of the tables in the specified 
' connection. That array is DESTRORED/REPLACED by the next call
' to the proceedure. This uses the sqlite3_exec()function so we
' actually have to write three functions. One of those will just 
' count the number of table names adn stores that value in 
' "SQL3TableCount". Then we execute the same call again but 
' redim and fill in the string array this time.
' These function are pretty much copied from the FreeBasic 
' exampled code translated by Nex/marzec.
'
Private Function GetTableCount cdecl ( _         
	byval NotUsed as any ptr, _             ' passed from FreeBasic call
	byval ColCount as integer, _            ' number of columns
	byval RowData as zstring ptr ptr, _     ' string array of data for one row
	byval ColName as zstring ptr ptr _      ' string array of column names
    ) as integer
    ' all we want to do here is count the number of rows
    ' we assume that the global variable has been zeroed
    SQL3TableCount=SQL3TableCount+1
    return 0
end function

Private Function GetTableStrings cdecl ( _
	byval NotUsed as any ptr, _             ' passed from FreeBasic call
	byval ColCount as integer, _            ' number of columns
	byval RowData as zstring ptr ptr, _     ' string array of data for one row
	byval ColName as zstring ptr ptr _      ' string array of column names
    ) as integer
    ' This time we have a little more work to do.
    ' We assume that the string array has already been resized
    ' and the SQL3TableNameCount set back to zero because we are
    ' going to use it for our index into the string array
    ' We only have one column to deal with and it 'should' not be empty
    ' NOTE: We start at one for our table names
    if( RowData[0] <> 0 ) then              ' check for null pointer
        if *RowData[0] <> 0 then            ' check for null pointer
            SQL3TableCount=SQL3TableCount+1    
            SQL3TableNames(SQL3TableCount) = *RowData[0]
        end if
    end if
    return 0
end function
    
Private Sub GetTableNamesPrv(FileNumber as Integer, Extended as Byte)
    ' This is the function that is actually does the work
    ' The second parameter controls exactly what query is used
    ' If zero then only user table names are returned
    ' If non-zero then all table names including internal are returned
    Dim X as Integer                        ' return code form SQLite
    Dim SQ as String                        ' our SQL query
    ' define our SQL String. 
    if Extended=0 then                      ' return only user table names
        SQ="SELECT name FROM sqlite_master WHERE type IN ('table','view') " & _
           "AND name NOT LIKE 'sqlite_%' UNION ALL SELECT name " & _ 
           "FROM sqlite_temp_master WHERE type IN ('table','view') " & _
           "ORDER BY 1;"                    ' note the terminating semicolon
    else                                    ' return all table names
        SQ="SELECT name FROM sqlite_master WHERE type IN ('table','view');" 
    end if
    'Print SQ 
    SQL3TableCount=0                        ' Zero our counter
    ' issue call to SQLite and count the number of rows
    X= sqlite3_exec( SQL3File(FileNumber).Handle, SQ, @GetTableCount, 0, @SQL3errMsg)
    ' redim our string array adn set counter back to zero
    Redim SQL3TableNames (0 to SQL3TableCount)
    SQL3TableCount=0
    ' issue call to SQLite and fill in the array
    X= sqlite3_exec( SQL3File(FileNumber).Handle, SQ, @GetTableStrings, 0, @SQL3errMsg)
end sub

Public Sub GetTableNamesUsr(FileNumber as Integer)
   ' fills in string array with user table names
   GetTableNamesPrv(FileNumber,0)
end Sub

Public Sub GetTableNamesAll(FileNumber as Integer)
   ' fills in string array with user table names
   GetTableNamesPrv(FileNumber,1)
end Sub
'------------------------------------------------------
' Now for column names --- Very Similar to  Table Names
' This function fills the arrays:
' "SQL3ColumnNames" with the names of the columns
' "SQL3ColumnTypes" with the column type
' "SQL3ColumnNull" with 0 if null values are not allowed.
' "SQL3ColumnDefault" with a string for the default value
' "SQL3ColumnPKey" with zeoe if not part of the primary key, otherwise the key index
' "SQL3ColumnCount" holds the count of the columns beginning at one
' These variables are DESTRORED/REPLACED by the next call to the proceedure.

Private Function GetColumnDataVal cdecl ( _
	byval NotUsed as any ptr, _             ' passed from FreeBasic call
	byval ColCount as integer, _            ' number of columns
	byval RowData as zstring ptr ptr, _     ' string array of data for one row
	byval ColName as zstring ptr ptr _      ' string array of column names
    ) as integer
    ' This time we have a little more work to do.
    ' We assume that the string array has already been resized
    ' and the SQL3ColumnCount set back to zero because we are
    ' going to use it for our index into the string arrays
    ' Field 0: the column number
    ' Field 1: the column name
    ' Field 2: the column type
    ' Field 3: the column null
    ' Field 4: the column default
    ' Field 5: the column Primary Key
    ' We are only interested in the first two fields: 
    '   column name
    '   column type
    ' NOTE: We start at one for our column names, types, etc.
    if( RowData[1] <> 0 ) then              ' check for null pointer
        if *RowData[1] <> 0 then            ' check for null pointer
            SQL3ColumnCount=SQL3ColumnCount+1    
            SQL3ColumnNames(SQL3ColumnCount) = *RowData[1]
            ' set the rest to an default empty string
            SQL3ColumnTypes(SQL3ColumnCount)=""
            SQL3ColumnNull(SQL3ColumnCount)=""
            SQL3ColumnDefault(SQL3ColumnCount)=""
            SQL3ColumnPKey(SQL3ColumnCount)=""
            ' now get the type
            if( RowData[2] <> 0 ) then
                if *RowData[2] <> 0 then
                    SQL3ColumnTypes(SQL3ColumnCount) =*RowData[2]
                end if
            end if
            ' now get null setting
            if( RowData[3] <> 0 ) then
                if *RowData[3] <> 0 then
                    SQL3ColumnNull(SQL3ColumnCount) =*RowData[3]
                end if
            end if
            ' now get default value
            if( RowData[4] <> 0 ) then
                if *RowData[4] <> 0 then
                    SQL3ColumnDefault(SQL3ColumnCount) =*RowData[4]
                end if
            end if
            ' now get primary key setting
            if( RowData[5] <> 0 ) then
                if *RowData[5] <> 0 then
                    SQL3ColumnPKey(SQL3ColumnCount) =*RowData[5]
                end if
            end if
        end if
    end if
    return 0
end function

Public Sub GetColumnData(FileNumber as Integer, TableName as String)
    ' This is the function that is actually does the work
    Dim SQ as String                        ' our SQL query
    ' define our SQL String. Note that future versions of SQLite may not
    ' support this particular PRAGMA .... then a more creative solution
    ' will be needed.
    SQ="PRAGMA table_info(" & TableName & ");"
    SQL3ColumnCount = SQL3QueryRowCount (FileNumber, SQ)
    ' redim our string array and set counter back to zero
    Redim SQL3ColumnNames   (0 to SQL3ColumnCount)
    Redim SQL3ColumnTypes   (0 to SQL3ColumnCount)
    Redim SQL3ColumnNull    (0 to SQL3ColumnCount)
    Redim SQL3ColumnDefault (0 to SQL3ColumnCount)
    Redim SQL3ColumnPkey    (0 to SQL3ColumnCount)
    SQL3ColumnCount=0
    ' issue call to SQLite and fill in the array
    SQL3errCode = sqlite3_exec( SQL3File(FileNumber).Handle, SQ, @GetColumnDataVal, 0, @SQL3errMsg)
end sub
'------------------------------------------------------
''  Return A tab delimited string for the Col Names for SQLQuery
Private Function SQL3QueryHeaderCallback cdecl ( _
	byval Hdr as any ptr, _              ' 
	byval ColCount as integer, _            ' number of columns
	byval RowData as zstring ptr ptr, _     ' string array of data for one row
	byval ColName as zstring ptr ptr _      ' string array of column names
    ) as integer
    ' This should only return one row with on column that contains the rowcount
    Dim Text as String=""
    Dim HText as String=""
    Dim Col as Integer
    For Col=0 to ColCount-1
        Text=""
        if( ColName[Col] <> 0 ) then              ' check for null pointer
            if *ColName[Col] <> 0 then            ' check for null pointer
                Text = Trim(*ColName[Col])
            end if
        end if
        HText=HText & Text & SQL3DLM
    Next Col
    Poke String, Hdr, HText
    return 0
end function

Public Function SQL3QueryHeader (FileNumber as Integer, Sql as String) as String
    ' returns the number of row for a SQL Query
    ' negative number indicates an error
    if SQL3File(FileNumber).Handle=0 then return "DB not open"
    Dim X as Integer
    Dim Hdr As String
    Dim Q as String
    Dim P as Integer
    SQL3errCode = sqlite3_exec( SQL3File(FileNumber).Handle, SQL, _ 
                     @SQL3QueryHeaderCallback, @Hdr, @SQL3errMsg)
    Return Hdr
End Function

''------------------------------------------------------
' Return A tab delimited string for the Data for SQLQuery
' Only use this for small queries against a know good datasource
' Can be use for returning information about the database
Private Function SQL3QueryDataCallback cdecl ( _
	byval RtnStr as any ptr, _              ' 
	byval ColCount as integer, _            ' number of columns
	byval RowData as zstring ptr ptr, _     ' string array of data for one row
	byval ColName as zstring ptr ptr _      ' string array of column names
    ) as integer
    ' This should only return one row with on column that contains the rowcount
    Dim Text as String=""
    Dim RText as String=""
    Dim Col as Integer
    ' build the row string
    For Col=0 to ColCount-1
        Text=""
        if( RowData[Col] <> 0 ) then              ' check for null pointer
            if *RowData[Col] <> 0 then            ' check for null pointer
                Text = Trim(*RowData[Col])
            end if
        end if
        RText=RText & Text & SQL3DLM
    Next Col
    ' Add the row to return string
    Poke String, RtnStr, Peek(String,RtnStr) & RText & SQL3EOL
    return 0
end function

Public Function SQL3QueryData (FileNumber as Integer, Sql as String) as String
    ' returns the number of row for a SQL Query
    ' negative number indicates an error
    if SQL3File(FileNumber).Handle=0 then return "DB not open"
    Dim X as Integer
    Dim RtnStr As String
    Dim Q as String
    Dim P as Integer
    SQL3errCode = sqlite3_exec( SQL3File(FileNumber).Handle, SQL, _ 
                     @SQL3QueryDataCallback, @RtnStr, @SQL3errMsg)
    Return RtnStr
End Function































